"""
This module implements a binary on/off switch, derived from the generic actor.
"""

from .generic2 import Generic2Actor


class SwitchActor(Generic2Actor):
    """A binary on/off switch actor."""

    name = "switch"
    config_defaults = {
        **Generic2Actor.config_defaults,
        "attributes": [{"attribute": "state"}],
        "values": [
            {"value": ["on"], "calls": [{"service": "homeassistant.turn_on"}]},
            {"value": ["off"], "calls": [{"service": "homeassistant.turn_off"}],},
        ],
        "ignore_case": True,
    }



"""
Module containing functionality to evaluate expressions.
"""

import types as _types
import typing as T

if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    from .. import schedule
    from ..room import Room

import datetime
import inspect

from . import helpers
from . import types  # pylint: disable=reimported


def build_expr_env(room: "Room", now: datetime.datetime) -> T.Dict[str, T.Any]:
    """This function builds and returns an environment usable as globals
    for the evaluation of an expression.
    It will add all members of the .types module's __all__ to the
    environment.
    Additionally, helpers provided by the .helpers module and the actor
    type will be constructed based on the Room object"""

    env = {}  # type: T.Dict[str, T.Any]
    for member_name in types.__all__:
        env[member_name] = getattr(types, member_name)

    helper_types = []
    for member_name, member in inspect.getmembers(helpers):
        if (
            member is not helpers.HelperBase
            and isinstance(member, type)
            and issubclass(member, helpers.HelperBase)
        ):
            helper_types.append(member)

    assert room.app.actor_type is not None
    helper_types.extend(room.app.actor_type.expression_helpers)

    helper_types.sort(key=lambda t: t.order)
    for helper_type in helper_types:
        room.log(
            "Initializing expression helper: {}, order = {}".format(
                helper_type.__name__, helper_type.order
            ),
            level="DEBUG",
        )
        helper = helper_type(room, now, env)
        helper.update_environment()

    return env


def eval_expr(expr: _types.CodeType, env: T.Dict[str, T.Any]) -> T.Any:
    """This method evaluates the given expression. The evaluation result
    is returned. The items of the env dict are added to the globals
    available during evaluation."""

    env = {**env}
    exec(expr, env)  # pylint: disable=exec-used
    return env.get("result")



"""
This module contains the CONFIG_SCHEMA for validation with voluptuous.
"""

import typing as T

import datetime
import traceback

import voluptuous as vol

from . import actor, schedule, util
from .room import Room


def build_schedule_rule(rule: dict) -> schedule.Rule:
    """Builds and returns a schedule rule from the given rule
    definition."""

    constraints = {}
    for name, value in rule.items():
        if name in schedule.Rule.CONSTRAINTS:
            constraints[name] = value

    expr = None
    expr_raw = rule.get("expression")
    if expr_raw is not None:
        expr_raw = expr_raw.strip()
        try:
            expr = util.compile_expression(expr_raw)
        except SyntaxError:
            traceback.print_exc(limit=0)
            raise vol.Invalid("Couldn't compile expression: {}".format(repr(expr_raw)))

    kwargs = {
        "start_time": rule["start"][0],
        "start_plus_days": rule["start"][1],
        "end_time": rule["end"][0],
        "end_plus_days": rule["end"][1],
        "constraints": constraints,
        "expr": expr,
        "expr_raw": expr_raw,
        "value": rule.get("value"),
    }

    if "rules" in rule:
        return schedule.SubScheduleRule(rule["rules"], **kwargs)
    return schedule.Rule(**kwargs)


def build_schedule(rules: T.Iterable[schedule.Rule]) -> schedule.Schedule:
    """Returns a Scheedule containing the given Rule objects."""

    sched = schedule.Schedule()
    for rule in rules:
        sched.rules.append(rule)
    return sched


def config_post_hook(cfg: dict) -> dict:
    """Creates Room and other objects after config has been parsed."""

    # pylint: disable=too-many-locals

    # name schedule snippets
    cfg["schedule_prepend"].name = "prepend"
    cfg["schedule_append"].name = "append"
    for name, sched in cfg["schedule_snippets"].items():
        sched.name = name

    actor_type = cfg["actor_type"]

    # Build Room objects.
    rooms = []
    for room_name, room_data in cfg["rooms"].items():
        actors = {}

        # copy defaults from templates and validate the actors
        for actor_name, actor_data in room_data["actors"].items():
            if "default" in cfg["actor_templates"]:
                actor_data.setdefault("template", "default")
            templates = [actor_data]
            while "template" in templates[-1]:
                template_name = templates[-1].pop("template")
                try:
                    template = cfg["actor_templates"][template_name].copy()
                except KeyError:
                    raise vol.ValueInvalid(
                        "No template named {} has been defined.".format(
                            repr(template_name)
                        )
                    )
                templates.append(template)

            _actor_data = {**actor_type.config_defaults}
            for template in reversed(templates):
                util.deep_merge_dicts(template, _actor_data)
            actor_data = vol.Schema(actor_type.config_schema_dict)(_actor_data)
            actors[actor_name] = actor_data

        # complete the room's schedule.
        rules = []
        if cfg["schedule_prepend"].rules:
            rules.append(schedule.SubScheduleRule(cfg["schedule_prepend"]))
        if room_data["schedule"].rules:
            room_data["schedule"].name = "room-individual"
            rules.append(schedule.SubScheduleRule(room_data["schedule"]))
        if cfg["schedule_append"].rules:
            rules.append(schedule.SubScheduleRule(cfg["schedule_append"]))
        sched = schedule.Schedule(name=room_name, rules=rules)

        del room_data["actors"]
        del room_data["schedule"]

        room = Room(room_name, room_data, cfg["_app"])
        rooms.append(room)

        # Create actor objects and attach to room.
        for actor_name, actor_data in actors.items():
            _actor = actor_type(actor_name, actor_data, room)
            room.actors.append(_actor)

        room.schedule = sched

    del cfg["rooms"], cfg["schedule_prepend"], cfg["schedule_append"]
    cfg["_app"].actor_type = actor_type
    cfg["_app"].rooms.extend(rooms)

    # Build StatisticalParameter objects.
    stats_params = []
    param_types = {t.name: t for t in actor_type.stats_param_types}
    for param_name, param_data in cfg["statistics"].items():
        param_data = STATISTICAL_PARAMETER_BASE_SCHEMA(param_data)
        param_type_name = param_data.pop("type")
        try:
            param_type = param_types[param_type_name]
        except KeyError:
            raise ValueError(
                "statistical parameter {} is not available for this actor type".format(
                    param_type_name
                )
            )

        _param_data = {**param_type.config_defaults}
        util.deep_merge_dicts(param_data, _param_data)
        param_data = vol.Schema(param_type.config_schema_dict)(_param_data)

        param = param_type(param_name, param_data, cfg["_app"])
        stats_params.append(param)

    cfg["_app"].stats_params.extend(stats_params)

    return cfg


def schedule_rule_pre_hook(rule: dict) -> dict:
    """Copy value for the expression and value keys over from alternative names."""

    rule = rule.copy()
    util.normalize_dict_key(rule, "expression", "x")
    util.normalize_dict_key(rule, "value", "v")
    return rule


def validate_rule_paths(sched: schedule.Schedule) -> schedule.Schedule:
    """A validator to be run after schedule creation to ensure
    each path contains at least one rule with an expression or value.
    A ValueError is raised when this check fails."""

    for path in sched.unfolded:
        if path.is_final and not list(path.rules_with_expr_or_value):
            raise ValueError(
                "No expression or value specified along the path {}.".format(path)
            )

    return sched


########## MISCELLANEOUS


def build_range_spec_validator(  # type: ignore
    min_value: int, max_value: int
) -> vol.Schema:
    """Returns a validator for range specifications with the given
    min/max values."""

    return vol.All(
        vol.Any(int, str), lambda v: util.expand_range_spec(v, min_value, max_value)
    )


ENTITY_ID_VALIDATOR = vol.Match(r"^[A-Za-z_]+\.[A-Za-z0-9_]+$")
PARTIAL_DATE_SCHEMA = vol.Schema(
    {
        vol.Optional("year"): vol.All(int, vol.Range(min=1970, max=2099)),
        vol.Optional("month"): vol.All(int, vol.Range(min=1, max=12)),
        vol.Optional("day"): vol.All(int, vol.Range(min=1, max=31)),
    }
)
TIME_VALIDATOR = vol.All(vol.Match(util.TIME_REGEXP), util.parse_time_string)
RULE_TIME_VALIDATOR = vol.All(
    vol.Match(
        util.RULE_TIME_REGEXP, msg="correct format: [<HH>:<MM>[:<SS>]][{+-}<days>d]"
    ),
    util.parse_rule_time_string,
)

# This schema does no real validation and default value insertion,
# it just ensures a dictionary containing string keys and dictionary
# values is given.
DICTS_IN_DICT_SCHEMA = vol.Schema(
    vol.All(lambda v: v or {}, {util.CONF_STR_KEY: vol.All(lambda v: v or {}, dict)})
)


def parse_watched_entity_str(value: str) -> T.Dict[str, T.Any]:
    """Parses the alternative <entity>:<attributes>:<mode> strings."""

    obj = {}  # type: T.Dict[str, T.Any]
    spl = [part.strip() for part in value.split(":")]
    if spl:
        part = spl.pop(0)
        if part:
            obj["entity"] = part
    if spl:
        part = spl.pop(0)
        if part:
            obj["attributes"] = [_part.strip() for _part in part.split(",")]
    if spl:
        part = spl.pop(0)
        if part:
            obj["mode"] = part
    return obj


WATCHED_ENTITY_SCHEMA = vol.Schema(
    vol.All(
        lambda v: v or {},
        vol.Any(vol.All(str, parse_watched_entity_str), object),
        {
            vol.Required("entity"): ENTITY_ID_VALIDATOR,
            vol.Optional("attributes", default="state"): vol.Any(
                vol.All(str, lambda v: [v]), [str]
            ),
            vol.Optional("mode", default="reevaluate"): vol.Any(
                "reevaluate", "reset", "ignore"
            ),
        },
    )
)
WATCHED_ENTITIES_SCHEMA = vol.Schema(
    vol.All(lambda v: v or [], [WATCHED_ENTITY_SCHEMA])
)


########## SCHEDULES

SCHEDULE_RULE_SCHEMA = vol.Schema(
    vol.All(
        lambda v: v or {},
        schedule_rule_pre_hook,
        {
            "rules": lambda v: SCHEDULE_SCHEMA(  # type: ignore  # pylint: disable=unnecessary-lambda
                v
            ),
            "expression": str,
            "value": object,
            vol.Optional("name", default=None): vol.Any(str, None),
            vol.Optional("start", default=(None, None)): vol.Any(
                RULE_TIME_VALIDATOR, (None,)
            ),
            vol.Optional("end", default=(None, None)): vol.Any(
                vol.All(
                    RULE_TIME_VALIDATOR,
                    (
                        None,
                        datetime.time,
                        vol.Range(min=0, msg="end time can't be shifted backwards"),
                    ),
                ),
                (None,),
            ),
            vol.Optional("years"): build_range_spec_validator(1970, 2099),
            vol.Optional("months"): build_range_spec_validator(1, 12),
            vol.Optional("days"): build_range_spec_validator(1, 31),
            vol.Optional("weeks"): build_range_spec_validator(1, 53),
            vol.Optional("weekdays"): build_range_spec_validator(1, 7),
            vol.Optional("start_date"): PARTIAL_DATE_SCHEMA,
            vol.Optional("end_date"): PARTIAL_DATE_SCHEMA,
        },
        build_schedule_rule,
    )
)

SCHEDULE_SCHEMA = vol.Schema(
    vol.All(lambda v: v or [], [SCHEDULE_RULE_SCHEMA], build_schedule)
)

SCHEDULE_SNIPPETS_SCHEMA = vol.Schema(
    vol.All(lambda v: v or {}, {util.CONF_STR_KEY: SCHEDULE_SCHEMA})
)


########## ROOMS

ROOM_SCHEMA = vol.Schema(
    vol.All(
        lambda v: v or {},
        {
            "friendly_name": str,
            vol.Optional("allow_manual_changes", default=True): bool,
            vol.Optional("replicate_changes", default=True): bool,
            vol.Optional("rescheduling_delay", default=0): vol.All(
                vol.Any(float, int), vol.Range(min=0)
            ),
            vol.Optional("actors", default=dict): DICTS_IN_DICT_SCHEMA,
            vol.Optional("watched_entities", default=None): WATCHED_ENTITIES_SCHEMA,
            vol.Optional("schedule", default=list): vol.All(
                SCHEDULE_SCHEMA, validate_rule_paths
            ),
        },
    )
)


########## STATISTICS

STATISTICAL_PARAMETER_BASE_SCHEMA = vol.Schema({vol.Required("type"): str}, extra=True)


########## MAIN CONFIG SCHEMA

CONFIG_SCHEMA = vol.Schema(
    vol.All(
        vol.Schema(
            {
                vol.Optional("reset_at_startup", default=False): bool,
                vol.Optional("expressions_from_events", default=False): bool,
                vol.Optional("expression_environment", default=None): vol.Any(
                    str, None
                ),
                vol.Required("actor_type"): vol.All(
                    vol.Any(*map(lambda a: a.name, actor.get_actor_types())),
                    lambda n: {a.name: a for a in actor.get_actor_types()}[n],
                ),
                vol.Optional("actor_templates", default=dict): DICTS_IN_DICT_SCHEMA,
                vol.Optional("watched_entities", default=dict): WATCHED_ENTITIES_SCHEMA,
                vol.Optional("schedule_prepend", default=list): vol.All(
                    SCHEDULE_SCHEMA, validate_rule_paths
                ),
                vol.Optional("schedule_append", default=list): vol.All(
                    SCHEDULE_SCHEMA, validate_rule_paths
                ),
                vol.Optional(
                    "schedule_snippets", default=dict
                ): SCHEDULE_SNIPPETS_SCHEMA,
                vol.Optional("rooms", default=dict): vol.All(
                    lambda v: v or {}, {util.CONF_STR_KEY: ROOM_SCHEMA}
                ),
                vol.Optional("statistics", default=dict): DICTS_IN_DICT_SCHEMA,
            },
            extra=True,
        ),
        config_post_hook,
    )
)



"""
This module implements the generic actor.
"""

import typing as T

import copy

import voluptuous as vol

from ... import common
from .base import ActorBase


ALLOWED_VALUE_TYPES = (bool, float, int, str, type(None))
ALLOWED_VALUE_TYPES_T = T.Union[  # pylint: disable=invalid-name
    bool, float, int, str, None
]
WILDCARD_ATTRIBUTE_VALUE = "*"


class Generic2Actor(ActorBase):
    """A configurable, generic actor for Schedy that can control multiple
    attributes at once."""

    name = "generic2"
    config_schema_dict = {
        **ActorBase.config_schema_dict,
        vol.Optional("attributes", default=None): vol.All(
            vol.DefaultTo(list),
            [
                vol.All(
                    vol.DefaultTo(dict),
                    {vol.Optional("attribute", default=None): vol.Any(str, None)},
                )
            ],
        ),
        vol.Optional("values", default=None): vol.All(
            vol.DefaultTo(list),
            [
                vol.All(
                    vol.DefaultTo(dict),
                    {
                        vol.Required("value"): vol.All(
                            [vol.Any(*ALLOWED_VALUE_TYPES)], vol.Coerce(tuple),
                        ),
                        vol.Optional("calls", default=None): vol.All(
                            vol.DefaultTo(list),
                            [
                                vol.All(
                                    vol.DefaultTo(dict),
                                    {
                                        vol.Required("service"): str,
                                        vol.Optional("data", default=None): vol.All(
                                            vol.DefaultTo(dict), dict
                                        ),
                                        vol.Optional(
                                            "include_entity_id", default=True
                                        ): bool,
                                    },
                                )
                            ],
                        ),
                    },
                )
            ],
            # Sort by number of attributes (descending) for longest prefix matching
            lambda v: sorted(v, key=lambda k: -len(k["value"])),
        ),
        vol.Optional("ignore_case", default=False): bool,
    }

    def _find_value_cfg(self, value: T.Tuple) -> T.Any:
        """Returns the config matching given value or ValueError if none found."""
        for value_cfg in self.cfg["values"]:
            _value = value_cfg["value"]
            if len(_value) != len(value):
                continue
            for idx, attr_value in enumerate(_value):
                if attr_value not in (WILDCARD_ATTRIBUTE_VALUE, value[idx]):
                    break
            else:
                return value_cfg
        raise ValueError("No configuration for value {!r}".format(value))

    def _populate_service_data(self, data: T.Dict, fmt: T.Dict[str, T.Any]) -> None:
        """Fills in placeholders in the service data definition."""
        # pylint: disable=too-many-nested-blocks
        memo = {data}  # type: T.Set[T.Union[T.Dict, T.List]]
        while memo:
            obj = memo.pop()
            if isinstance(obj, dict):
                _iter = obj.items()  # type: T.Iterable[T.Tuple[T.Any, T.Any]]
            elif isinstance(obj, list):
                _iter = enumerate(obj)
            else:
                continue
            for key, value in _iter:
                if isinstance(value, str):
                    try:
                        formatted = value.format(fmt)
                        # Convert special values to appropriate type
                        if formatted == "None":
                            obj[key] = None
                        elif formatted == "True":
                            obj[key] = True
                        elif formatted == "False":
                            obj[key] = False
                        else:
                            try:
                                _float = float(formatted)
                                _int = int(formatted)
                            except ValueError:
                                # It's a string value
                                obj[key] = formatted
                            else:
                                # It's an int or float
                                obj[key] = _int if _float == _int else _float
                    except (IndexError, KeyError, ValueError) as err:
                        self.log(
                            "Couldn't format service data {!r} with values "
                            "{!r}: {!r}, omitting data.".format(value, fmt, err),
                            level="ERROR",
                        )
                elif isinstance(value, (dict, list)):
                    memo.add(value)

    def do_send(self) -> None:
        """Executes the configured services for self._wanted_value."""
        value = self._wanted_value
        # Build formatting data with values of all attributes
        fmt = {"entity_id": self.entity_id}
        for idx in range(len(self.cfg["attributes"])):
            fmt["attr{}".format(idx + 1)] = value[idx] if idx < len(value) else None

        for call_cfg in self._find_value_cfg(value)["calls"]:
            service = call_cfg["service"]
            data = copy.deepcopy(call_cfg["data"])
            self._populate_service_data(data, fmt)
            if call_cfg["include_entity_id"]:
                data.setdefault("entity_id", self.entity_id)
            self.log(
                "Calling service {}, data = {}.".format(repr(service), repr(data)),
                level="DEBUG",
                prefix=common.LOG_PREFIX_OUTGOING,
            )
            self.app.call_service(service, **data)

    def filter_set_value(self, value: T.Tuple) -> T.Any:
        """Checks whether the actor supports this value."""
        if self.cfg["ignore_case"]:
            value = tuple(v.lower() if isinstance(v, str) else v for v in value)
        try:
            self._find_value_cfg(value)
        except ValueError:
            self.log(
                "Value {!r} is not known by this actor.".format(value), level="ERROR"
            )
            return None
        return value

    def notify_state_changed(self, attrs: dict) -> T.Any:
        """Is called when the entity's state changes."""
        items = []
        for attr_cfg in self.cfg["attributes"]:
            attr = attr_cfg["attribute"]
            if attr is None:
                self.log("Ignoring state change (write-only attribute).", level="DEBUG")
                return None
            state = attrs.get(attr)
            self.log(
                "Attribute {!r} is {!r}.".format(attr, state),
                level="DEBUG",
                prefix=common.LOG_PREFIX_INCOMING,
            )
            if self.cfg["ignore_case"] and isinstance(state, str):
                state = state.lower()
            items.append(state)

        tpl = tuple(items)
        # Goes from len(tpl) down to 0
        for size in range(len(tpl), -1, -1):
            value = tpl[:size]
            try:
                self._find_value_cfg(value)
            except ValueError:
                continue
            return value

        self.log(
            "Received state {!r} which is not configured as a value.".format(items),
            level="WARNING",
        )
        return None

    @staticmethod
    def validate_value(value: T.Any) -> T.Any:
        """Converts lists to tuples."""
        if isinstance(value, list):
            items = tuple(value)
        elif isinstance(value, tuple):
            items = value
        else:
            items = (value,)

        for index, item in enumerate(items):
            if not isinstance(item, ALLOWED_VALUE_TYPES):
                raise ValueError(
                    "Value {!r} for {}. attribute must be of one of these types: "
                    "{}".format(item, index + 1, ALLOWED_VALUE_TYPES)
                )
        return items



"""
This module implements the thermostat actor.
"""

import typing as T

if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    from ..room import Room

import functools
import voluptuous as vol

from ... import common
from .. import stats
from ..expression.helpers import HelperBase as ExpressionHelperBase
from .base import ActorBase


# allowed types of values to initialize Temp() with
TempValueType = T.Union[float, int, str, "Off", "Temp"]


class Off:
    """A special value Temp() may be initialized with in order to turn
    a thermostat off."""

    def __add__(self, other: T.Any) -> "Off":
        return self

    def __eq__(self, other: T.Any) -> bool:
        return isinstance(other, Off)

    def __hash__(self) -> int:
        return hash(str(self))

    def __neg__(self) -> "Off":
        return self

    def __repr__(self) -> str:
        return "OFF"

    def __sub__(self, other: T.Any) -> "Off":
        return self


OFF = Off()


@functools.total_ordering
class Temp:
    """A class holding a temperature value."""

    def __init__(self, temp_value: T.Any) -> None:
        if isinstance(temp_value, Temp):
            # Just copy the value over.
            parsed = self.parse_temp(temp_value.value)
        else:
            parsed = self.parse_temp(temp_value)

        if parsed is None:
            raise ValueError("{} is no valid temperature".format(repr(temp_value)))

        self.value = parsed  # type: T.Union[float, Off]

    def __add__(self, other: T.Any) -> "Temp":
        if isinstance(other, (float, int)):
            other = type(self)(other)
        elif not isinstance(other, type(self)):
            return NotImplemented

        # OFF + something is OFF
        if self.is_off or other.is_off:
            return type(self)(OFF)
        return type(self)(self.value + other.value)

    def __eq__(self, other: T.Any) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value == other.value

    def __float__(self) -> float:
        if isinstance(self.value, float):
            return self.value
        raise ValueError("{} has no numeric value.".format(repr(self)))

    def __hash__(self) -> int:
        return hash(str(self))

    def __lt__(self, other: T.Any) -> bool:
        if isinstance(other, (float, int)):
            other = type(self)(other)
        if type(self) is not type(other):
            return NotImplemented
        if not self.is_off and other.is_off:
            return False
        if self.is_off and not other.is_off or self.value < other.value:
            return True
        return False

    def __neg__(self) -> "Temp":
        return type(self)(-self.value)  # pylint: disable=invalid-unary-operand-type

    def __repr__(self) -> str:
        if isinstance(self.value, (float, int)):
            return "{}°".format(self.value)
        return "{}".format(self.value)

    def __sub__(self, other: T.Any) -> "Temp":
        return self.__add__(-other)

    @property
    def is_off(self) -> bool:
        """Tells whether this temperature means OFF."""

        return isinstance(self.value, Off)

    @staticmethod
    def parse_temp(value: T.Any) -> T.Union[float, Off, None]:
        """Converts the given value to a valid temperature of type float
        or Off.
        If value is a string, all whitespace is removed first.
        If conversion is not possible, None is returned."""

        if isinstance(value, str):
            value = "".join(value.split())
            if value.upper() == "OFF":
                return OFF

        if isinstance(value, Off):
            return OFF

        try:
            return float(value)
        except (ValueError, TypeError):
            return None

    def serialize(self) -> str:
        """Converts the temperature into a string that Temp can be
        initialized with again later."""

        if self.is_off:
            return "OFF"
        return str(self.value)


class ThermostatExpressionHelper(ExpressionHelperBase):
    """Adds Temp and OFF to the evaluation environment."""

    OFF = OFF
    Temp = Temp


TEMP_SCHEMA = vol.Schema(
    vol.All(
        vol.Any(float, int, Off, vol.All(str, lambda v: v.upper(), "OFF")),
        lambda v: Temp(v),  # pylint: disable=unnecessary-lambda
    )
)


class TempDeltaParameter(stats.ActorValueCollectorMixin, stats.MinAvgMaxParameter):
    """The difference between target and current temperature."""

    name = "temp_delta"
    config_schema_dict = {
        **stats.ActorValueCollectorMixin.config_schema_dict,
        **stats.MinAvgMaxParameter.config_schema_dict,
        vol.Optional("off_value", default=0): vol.Any(float, int, None),
    }
    round_places = 2

    def collect_actor_value(self, actor: ActorBase) -> T.Optional[float]:
        """Collects the difference between target and current temperature."""

        assert isinstance(actor, ThermostatActor)
        current = actor.current_temp
        target = actor.current_value
        if current is None or target is None or current.is_off or target.is_off:
            off_value = self.cfg["off_value"]
            if off_value is None:
                # thermostats that are off should be excluded
                return None
            return float(off_value)
        return float(target - current)

    def initialize_actor_listeners(self, actor: ActorBase) -> None:
        """Listens for changes of current and target temperature."""

        self.log(
            "Listening for temperature changes of {} in {}.".format(actor, actor.room),
            level="DEBUG",
        )
        actor.events.on("current_temp_changed", self.update_handler)
        actor.events.on("value_changed", self.update_handler)


class ThermostatActor(ActorBase):
    """A thermostat to be controlled by Schedy."""

    name = "thermostat"
    config_schema_dict = {
        **ActorBase.config_schema_dict,
        vol.Optional("delta", default=0): vol.All(TEMP_SCHEMA, vol.NotIn([Temp(OFF)])),
        vol.Optional("min_temp", default=None): vol.Any(
            vol.All(TEMP_SCHEMA, vol.NotIn([Temp(OFF)])), None
        ),
        vol.Optional("max_temp", default=None): vol.Any(
            vol.All(TEMP_SCHEMA, vol.NotIn([Temp(OFF)])), None
        ),
        vol.Optional("off_temp", default=OFF): TEMP_SCHEMA,
        vol.Optional("supports_hvac_modes", default=True): bool,
        vol.Optional("hvac_mode_on", default="heat"): str,
        vol.Optional("hvac_mode_off", default="off"): str,
    }

    expression_helpers = ActorBase.expression_helpers + [ThermostatExpressionHelper]

    stats_param_types = [TempDeltaParameter]

    def __init__(self, *args: T.Any, **kwargs: T.Any) -> None:
        super().__init__(*args, **kwargs)
        self._current_temp = None  # type: T.Optional[Temp]

    def check_config_plausibility(self, state: dict) -> None:
        """Is called during initialization to warn the user about some
        possible common configuration mistakes."""

        if not state:
            self.log("Thermostat couldn't be found.", level="WARNING")
            return

        required_attrs = ["temperature"]
        if self.cfg["supports_hvac_modes"]:
            required_attrs.append("state")
        for attr in required_attrs:
            if attr not in state:
                self.log(
                    "Thermostat has no attribute named {!r}. Available are {!r}. "
                    "Please check your config!".format(attr, list(state.keys())),
                    level="WARNING",
                )

        temp_attrs = ["temperature", "current_temperature"]
        for attr in temp_attrs:
            value = state.get(attr)
            try:
                value = float(value)  # type: ignore
            except (TypeError, ValueError):
                self.log(
                    "The value {!r} of attribute {!r} is no valid temperature.".format(
                        value, attr
                    ),
                    level="WARNING",
                )

        allowed_hvac_modes = state.get("hvac_modes")
        if not self.cfg["supports_hvac_modes"]:
            if allowed_hvac_modes:
                self.log(
                    "HVAC mode support has been disabled, but the modes {!r} seem to "
                    "be supported. Maybe disabling it was a mistake?".format(
                        allowed_hvac_modes
                    ),
                    level="WARNING",
                )
            return

        if not allowed_hvac_modes:
            self.log(
                "Attributes for thermostat contain no 'hvac_modes', Consider "
                "disabling HVAC mode support.",
                level="WARNING",
            )
            return
        for hvac_mode in (self.cfg["hvac_mode_on"], self.cfg["hvac_mode_off"]):
            if hvac_mode not in allowed_hvac_modes:
                self.log(
                    "Thermostat doesn't seem to support the "
                    "HVAC mode {}, supported modes are: {}. "
                    "Please check your config!".format(hvac_mode, allowed_hvac_modes),
                    level="WARNING",
                )

    @property
    def current_temp(self) -> T.Optional[Temp]:
        """Returns the current temperature as measured by the thermostat."""

        return self._current_temp

    @staticmethod
    def deserialize_value(value: str) -> Temp:
        """Deserializes by calling validate_value()."""

        return ThermostatActor.validate_value(value)

    def do_send(self) -> None:
        """Sends self._wanted_value to the thermostat."""

        target_temp = self._wanted_value
        if target_temp.is_off:
            hvac_mode = self.cfg["hvac_mode_off"]
            temp = None
        else:
            hvac_mode = self.cfg["hvac_mode_on"]
            temp = target_temp
        if not self.cfg["supports_hvac_modes"]:
            hvac_mode = None

        self.log(
            "Setting temperature = {!r}, HVAC mode = {!r}.".format(
                "<unset>" if temp is None else temp,
                "<unset>" if hvac_mode is None else hvac_mode,
            ),
            level="DEBUG",
            prefix=common.LOG_PREFIX_OUTGOING,
        )
        if hvac_mode is not None:
            self.app.call_service(
                "climate/set_hvac_mode", entity_id=self.entity_id, hvac_mode=hvac_mode
            )
        if temp is not None:
            self.app.call_service(
                "climate/set_temperature",
                entity_id=self.entity_id,
                temperature=temp.value,
            )

    def filter_set_value(self, value: Temp) -> T.Optional[Temp]:
        """Preprocesses the given target temperature for setting on this
        thermostat. This algorithm will try best to achieve the closest
        possible temperature supported by this particular thermostat.
        The return value is either the temperature to set or None,
        if nothing has to be sent."""

        if value.is_off:
            value = self.cfg["off_temp"]

        if not value.is_off:
            value = value + self.cfg["delta"]
            if isinstance(self.cfg["min_temp"], Temp) and value < self.cfg["min_temp"]:
                value = self.cfg["min_temp"]
            elif (
                isinstance(self.cfg["max_temp"], Temp) and value > self.cfg["max_temp"]
            ):
                value = self.cfg["max_temp"]
        elif not self.cfg["supports_hvac_modes"]:
            self.log(
                "Not turning off because it doesn't support HVAC modes.",
                level="WARNING",
            )
            self.log(
                "Consider defining an off_temp in the actor "
                "configuration for these cases.",
                level="WARNING",
            )
            return None

        return value

    def notify_state_changed(self, attrs: dict) -> T.Any:
        """Is called when the thermostat's state changes.
        This method fetches both the current and target temperature from
        the thermostat and reacts accordingly."""

        _target_temp = None  # type: T.Optional[TempValueType]
        if self.cfg["supports_hvac_modes"]:
            hvac_mode = attrs.get("state")
            self.log(
                "Attribute 'state' is {}.".format(repr(hvac_mode)),
                level="DEBUG",
                prefix=common.LOG_PREFIX_INCOMING,
            )
            if hvac_mode == self.cfg["hvac_mode_off"]:
                _target_temp = OFF
            elif hvac_mode != self.cfg["hvac_mode_on"]:
                self.log(
                    "Unknown HVAC mode {!r}, ignoring thermostat.".format(hvac_mode),
                    level="ERROR",
                )
                return None
        else:
            hvac_mode = None

        if _target_temp is None:
            _target_temp = attrs.get("temperature")
            self.log(
                "Attribute 'temperature' is {}.".format(repr(_target_temp)),
                level="DEBUG",
                prefix=common.LOG_PREFIX_INCOMING,
            )

        try:
            target_temp = Temp(_target_temp)
        except ValueError:
            self.log("Invalid target temperature, ignoring thermostat.", level="ERROR")
            return None

        _current_temp = attrs.get("current_temperature")
        self.log(
            "Attribute 'current_temperature' is {}.".format(repr(_current_temp)),
            level="DEBUG",
            prefix=common.LOG_PREFIX_INCOMING,
        )
        if _current_temp is not None:
            try:
                current_temp = Temp(_current_temp)  # type: T.Optional[Temp]
            except ValueError:
                self.log(
                    "Invalid current temperature {!r}, not updating it.".format(
                        _current_temp
                    ),
                    level="ERROR",
                )
            else:
                if current_temp != self._current_temp:
                    self._current_temp = current_temp
                    self.events.trigger("current_temp_changed", self, current_temp)

        return target_temp

    @staticmethod
    def serialize_value(value: Temp) -> str:
        """Wrapper around Temp.serialize()."""

        if not isinstance(value, Temp):
            raise ValueError(
                "can only serialize Temp objects, not {}".format(repr(value))
            )
        return value.serialize()

    @staticmethod
    def validate_value(value: T.Any) -> Temp:
        """Ensures the given value is a valid temperature."""

        return Temp(value)



"""
This package contains the various actor implementations.
"""

import typing as T

from .base import ActorBase
from .custom import CustomActor
from .generic import GenericActor
from .generic2 import Generic2Actor
from .switch import SwitchActor
from .thermostat import ThermostatActor


__all__ = [
    "ActorBase",
    "CustomActor",
    "GenericActor",
    "Generic2Actor",
    "SwitchActor",
    "ThermostatActor",
]


def get_actor_types() -> T.Iterable[T.Type[ActorBase]]:
    """Yields available actor classes."""

    globs = globals()
    for actor_class_name in __all__:
        actor_type = globs.get(actor_class_name)
        if (
            actor_type is not ActorBase
            and isinstance(actor_type, type)
            and issubclass(actor_type, ActorBase)
        ):
            yield actor_type



"""
This module implements the custom actor.
"""

import types
import typing as T

import voluptuous as vol

from .. import util
from .base import ActorBase


class CustomActor(ActorBase):
    """A fully customizable actor for Schedy."""

    name = "custom"
    config_schema_dict = {
        **ActorBase.config_schema_dict,
        vol.Required("send_hook"): vol.All(str, util.compile_expression),
        "state_hook": vol.All(str, util.compile_expression),
        "filter_value_hook": vol.All(str, util.compile_expression),
        vol.Optional("config", default=dict): vol.All(lambda v: v or {}, dict),
    }

    def _exec_script(self, expr: types.CodeType, env: T.Dict[str, T.Any]) -> T.Any:
        """evaluates the given expression and returns the value of
        result or None, if unset or an error occured."""

        env = {**env}
        env.setdefault("entity_id", self.entity_id)
        env.setdefault("config", self.cfg["config"])
        env.setdefault("app", self.app)
        env.setdefault("actor", self)

        try:
            exec(expr, env)  # pylint: disable=exec-used
        except Exception as err:  # pylint: disable=broad-except
            self.log(
                "Error while evaluating expression: {}".format(repr(err)), level="ERROR"
            )
            return None
        return env.get("result")

    def do_send(self) -> None:
        """Executes the configured send script for self._wanted_value."""

        self.log("Executing send script.", level="DEBUG")
        env = {"value": self._wanted_value}
        self._exec_script(self.cfg["send_hook"], env)

    def filter_set_value(self, value: T.Any) -> T.Any:
        """Executes the configured filter_value script."""

        if "filter_value_hook" in self.cfg:
            env = {"value": value}
            result = self._exec_script(self.cfg["filter_value_hook"], env)
            self.log(
                "Filter rewrote value {} to {}.".format(repr(value), repr(result)),
                level="WARNING",
            )
            return result
        return value

    def notify_state_changed(self, attrs: dict) -> T.Any:
        """Is called when the entity's state changes."""

        if "state_hook" not in self.cfg:
            return None

        env = {"state": attrs}
        value = self._exec_script(self.cfg["state_hook"], env)
        self.log(
            "State {} resulted in a value of {}.".format(repr(attrs), repr(value)),
            level="DEBUG",
        )
        if value is None:
            self.log("Ignoring value of None.", level="DEBUG")
            return None
        return value



#!/usr/bin/env python


import os
from setuptools import find_packages, setup

from hass_apps import __version__


def read_file(filename):
    """Returns content of the given file. Path must be relative to
    this file."""

    with open(os.path.join(os.path.dirname(__file__), filename)) as file:
        return file.read()


setup(
    name = "hass-apps",
    version = __version__,
    description = "A collection of useful apps for AppDaemon + "
                  "Home Assistant.",
    long_description = read_file("README.rst"),
    author = "Robert Schindler",
    author_email = "r.schindler@efficiosoft.com",
    url = "https://github.com/efficiosoft/hass-apps",
    license = "Apache 2.0",
    packages = find_packages("."),
    install_requires = [
        "appdaemon >= 3.0",
        "cached-property >= 1.5.1",
        "observable >= 1.0",
        "voluptuous >= 0.11",
    ],
    zip_safe = False,
)



"""
This module implements the ActorBase parent class.
"""

import typing as T

if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    import uuid
    from ..expression.helpers import HelperBase as ExpressionHelperBase
    from ..room import Room
    from ..stats import StatisticalParameter

import copy
import json
import observable
import voluptuous as vol

from ... import common
from ..room import sync_proxy


class ActorBase:
    """An actor to be controlled by Schedy."""

    name = "actor"
    config_defaults = {}  # type: T.Dict[T.Any, T.Any]
    config_schema_dict = {
        "friendly_name": str,
        vol.Optional("send_retries", default=10): vol.All(int, vol.Range(min=0)),
        vol.Optional("send_retry_interval", default=30): vol.All(int, vol.Range(min=1)),
    }

    expression_helpers = []  # type: T.List[T.Type[ExpressionHelperBase]]

    stats_param_types = []  # type: T.List[T.Type[StatisticalParameter]]

    def __init__(self, entity_id: str, cfg: dict, room: "Room") -> None:
        self.entity_id = entity_id
        self.cfg = cfg
        self.room = room
        self.app = room.app
        self.events = observable.Observable()  # type: observable.Observable
        self.is_initialized = False

        self._current_value = None  # type: T.Any
        self._wanted_value = None  # type: T.Any

        self._gave_up_sending = False
        self._resending_timer = None  # type: T.Optional[uuid.UUID]

    def __repr__(self) -> str:
        return "<Actor {}>".format(str(self))

    def __str__(self) -> str:
        return "A:{}".format(self.cfg.get("friendly_name", self.entity_id))

    @staticmethod
    def _preprocess_state(state: T.Optional[dict]) -> dict:
        """Copies and flattens a state dict."""

        attrs = copy.deepcopy(state or {})
        attrs.update((attrs or {}).get("attributes", {}))
        return attrs

    @sync_proxy
    def _resending_cb(self, kwargs: dict) -> None:
        """This callback triggers the actual sending of a value to the
        actor. Expected members of kwargs are:
        - left_tries (including this round)"""

        self._resending_timer = None
        retries = self.cfg["send_retries"]
        interval = self.cfg["send_retry_interval"]
        left_tries = kwargs["left_tries"]
        if not left_tries:
            self.log(
                "Gave up sending value after {} retries.".format(retries),
                level="WARNING" if retries else "DEBUG",
            )
            self._gave_up_sending = True
            return
        if left_tries <= retries:
            self.log("Re-sending value due to missing confirmation.", level="WARNING")

        self.log(
            "Setting value {!r} (left tries = {}, interval = {}).".format(
                self._wanted_value, left_tries, interval
            ),
            level="DEBUG",
            prefix=common.LOG_PREFIX_OUTGOING,
        )
        self.do_send()

        self._gave_up_sending = False
        self._resending_timer = self.app.run_in(
            self._resending_cb, interval, left_tries=left_tries - 1
        )

    @sync_proxy
    def _state_cb(
        self,
        entity: str,
        attr: str,
        old: T.Optional[dict],
        new: T.Optional[dict],
        kwargs: dict,
    ) -> None:
        """Is called when any of the actor's state attributes changes."""

        attrs = self._preprocess_state(new)

        previous_value = self._current_value
        new_value = self.notify_state_changed(  # pylint: disable=assignment-from-none
            attrs
        )
        if new_value is None:
            return

        if new_value == self._wanted_value:
            self.cancel_resending_timer()
            self._gave_up_sending = False

        if new_value != previous_value:
            self._current_value = new_value
            self.log(
                "Received value of {}.".format(repr(new_value)),
                prefix=common.LOG_PREFIX_INCOMING,
            )
            self.events.trigger("value_changed", self, new_value)

    def after_initialization(self) -> None:
        """Can be implemented to perform actions after actor initialization."""

    def cancel_resending_timer(self) -> None:
        """Cancels the re-sending timer for this actor, if one exists."""

        timer = self._resending_timer
        if timer is None:
            return
        self._resending_timer = None
        self.app.cancel_timer(timer)
        self.log("Cancelled re-sending timer.", level="DEBUG")

    def check_config_plausibility(self, state: dict) -> None:
        """Is called during initialization to warn the user about some
        possible common configuration mistakes. The entity's current
        state attributes dictionary is provided."""

    @property
    def current_value(self) -> T.Any:
        """Returns the value currently set at this actor."""

        return self._current_value

    @staticmethod
    def deserialize_value(value: str) -> T.Any:
        """Should deserialize a value generated by serialize_value().
        A ValueError should be raised in case of malformed data.
        This implementation uses JSON."""

        try:
            return json.loads(value)
        except (TypeError, json.JSONDecodeError) as err:
            raise ValueError("invalid JSON data: {}".format(repr(err)))

    def do_send(self) -> None:
        """This method should implement the actual sending of
        self._wanted_value to the actor."""

    def filter_set_value(self, value: T.Any) -> T.Any:  # pylint: disable=no-self-use
        """Should be implemented to decide whether to set the given
        value on this actor or to alter it before setting.
        The return value is either the actual value to set or None,
        if nothing should be sent."""

        return value

    @property
    def gave_up_sending(self) -> bool:
        """Tells whether actor gave up sending and is still waiting for a receipt."""

        return self._gave_up_sending

    @property
    def is_sending(self) -> bool:
        """Tells whether the actor is currently waiting for a receipt."""

        return self._resending_timer is not None

    @property
    def is_synced(self) -> bool:
        """Tells whether the actor's current value is the wanted one and
        re-sending neither is in progress nor has failed."""

        return (
            not self._resending_timer
            and not self._gave_up_sending
            and self._current_value is not None
            and self._wanted_value is not None
            and self._current_value == self._wanted_value
        )

    def log(self, msg: str, *args: T.Any, **kwargs: T.Any) -> None:
        """Prefixes the actor to log messages."""

        msg = "[{}] {}".format(self, msg)
        self.room.log(msg, *args, **kwargs)

    def notify_state_changed(  # pylint: disable=no-self-use,unused-argument
        self, attrs: dict
    ) -> T.Any:
        """Is called when the entity's state has changed with the new
        attributes dict as argument. It should return the new value or
        None, if undetectable."""

        return None

    def initialize(self) -> bool:
        """Should be called in order to register state listeners and
        timers.
        Returns whether initialization was successful."""

        self.log(
            "Initializing actor (entity_id={}, type={}).".format(
                repr(self.entity_id), repr(self.name)
            ),
            level="DEBUG",
        )

        self.log("Fetching initial state.", level="DEBUG")
        state = self.app.get_state(self.entity_id, attribute="all")
        if state is None:
            self.log(
                "State of entity {} is None, not initializing it now.".format(
                    repr(self.entity_id)
                ),
                level="WARNING",
            )
            return False
        self.check_config_plausibility(self._preprocess_state(state))
        # populate self._current_value etc. by simulating a
        # state change
        self._state_cb(self.entity_id, "all", state, state, {})

        self.log("Listening for state changes.", level="DEBUG")
        self.app.listen_state(self._state_cb, self.entity_id, attribute="all")

        self.after_initialization()

        self.is_initialized = True
        return True

    @staticmethod
    def serialize_value(value: T.Any) -> str:
        """Should serialize accepted values for this actor to str.
        A ValueError should be raised in case of not serializable data.
        This implementation uses JSON."""

        try:
            return json.dumps(value)
        except TypeError as err:
            raise ValueError("can't serialize to JSON: {}".format(err))

    def set_value(
        self, value: T.Any, force_resend: bool = False
    ) -> T.Tuple[bool, T.Any]:
        """Is called in order to change the actor's value. It isn't
        re-sent unless force_resend is True.
        It returns whether a value has been sent or not and the actual
        value now wanted by this actor."""

        value = self.filter_set_value(value)
        if value is None:
            return False, self._wanted_value

        self._wanted_value = value
        if not force_resend and self.is_synced:
            self.log(
                "Not sending value {} redundantly.".format(repr(value)), level="DEBUG"
            )
            return False, value

        self.cancel_resending_timer()
        self._resending_cb({"left_tries": self.cfg["send_retries"] + 1})

        return True, value

    @staticmethod
    def validate_value(value: T.Any) -> T.Any:
        """Should validate the given value, which usually is the result
        of a custom expression, to ensure it's appropriate for this kind
        of actor. It may alter the value before returning it again.
        A ValueError should be raised when validation fails."""

        return value

    @property
    def wanted_value(self) -> T.Any:
        """Returns the value currently wanted for this actor."""

        return self._wanted_value

    @wanted_value.setter
    def wanted_value(self, value: T.Any) -> None:
        """Validates and manually sets the value wanted by this actor."""

        if value is not None:
            value = self.validate_value(value)
        self._wanted_value = value



"""
This module implements various classes for collecting statistics.
"""

import typing as T

if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    import uuid
    from .app import SchedyApp
    from .room import Room
    from .actor.base import ActorBase

import voluptuous as vol

from .. import common
from . import util


StatisticalValueType = T.Union[float, int, str]


class StatisticalParameter:
    """A parameter to be collected."""

    name = "dummy"
    config_defaults = {}  # type: T.Dict[T.Any, T.Any]
    config_schema_dict = {}  # type: T.Dict[T.Any, T.Any]

    # number of decimal places to round numeric values to,
    # None disables rounding
    round_places = None  # type: T.Optional[int]

    def __init__(self, name: str, cfg: T.Dict, app: "SchedyApp") -> None:
        self.name = name
        self.cfg = cfg
        self.app = app

        self.rooms = []  # type: T.List[Room]

        self._last_state = None  # type: T.Optional[T.Dict[str, StatisticalValueType]]
        self._update_timer = None  # type: T.Optional[uuid.UUID]

    def __repr__(self) -> str:
        return "<StatisticalParameter {}>".format(self.name)

    def __str__(self) -> str:
        return "SP:{}".format(self.name)

    def generate_entries(  # pylint: disable=no-self-use
        self,
    ) -> T.Dict[str, StatisticalValueType]:
        """Should generate the entries to be added to the parameter."""

        return {}

    def _do_update(self) -> None:
        """Collects the statistics and writes them to Home Assistant."""

        self._update_timer = None

        attrs = self.generate_entries()

        if self.round_places is not None:
            for attr, value in attrs.items():
                if isinstance(value, (float, int)):
                    attrs[attr] = util.round_number(value, self.round_places)

        unchanged = attrs == self._last_state
        if unchanged:
            self.log("Unchanged HA state: attributes={}".format(attrs), level="DEBUG")
            return
        self.log(
            "Sending state to HA: attributes={}".format(attrs),
            level="DEBUG",
            prefix=common.LOG_PREFIX_OUTGOING,
        )

        entity_id = self._state_entity_id
        self.app.set_state(entity_id, state="", attributes=attrs)
        self._last_state = attrs

    @property
    def _state_entity_id(self) -> str:
        """Generates the entity id for storing this parameter's state as."""

        return "schedy_stats.{}_{}".format(self.app.name, self.name)

    def initialize(self) -> None:
        """Initializes state listeners and triggers an initial update."""

        self.log(
            "Initializing statistical parameter (name={}).".format(repr(self.name)),
            level="DEBUG",
        )

        self.initialize_listeners()

        self.update()

    def initialize_listeners(self) -> None:  # pylint: disable=no-self-use
        """Is called during initialization and should set up
        implementation-specific state listeners."""

    def log(self, msg: str, *args: T.Any, **kwargs: T.Any) -> None:
        """Prefixes the parameter to log messages."""

        msg = "[{}] {}".format(self, msg)
        self.app.log(msg, *args, **kwargs)

    def update(self) -> None:
        """Registers a timer for sending statistics to HA in 3 seconds."""

        if self._update_timer:
            self.log("Statistics update  pending already.", level="DEBUG")
            return

        self.log("Going to update statistics in 3 seconds.", level="DEBUG")
        self._update_timer = self.app.run_in(lambda *a, **kw: self._do_update(), 3)

    def update_handler(
        self, *args: T.Any, **kwargs: T.Any  # pylint: disable=unused-argument
    ) -> None:
        """A convenience wrapper around self.update(), accepting any
        positional and wildcard argument. It is intended to be used
        with events that pass arguments to their handler we're not
        interested in."""

        self.update()


class RoomBasedParameter(StatisticalParameter):
    """A parameter having rooms associated."""

    config_schema_dict = {
        **StatisticalParameter.config_schema_dict,
        vol.Optional("rooms", default=dict): vol.All(
            lambda v: v or {},
            {
                util.CONF_STR_KEY: vol.Schema(
                    vol.All(
                        lambda v: v or {},
                        {
                            # no settings yet
                        },
                    )
                )
            },
        ),
    }

    def __init__(self, *args: T.Any, **kwargs: T.Any) -> None:
        super().__init__(*args, **kwargs)
        self.rooms = []  # type: T.List[Room]

    def initialize_listeners(self) -> None:
        """Adds all configured rooms to self.rooms."""

        for room_name in self.cfg["rooms"]:
            for room in self.app.rooms:
                if room.name == room_name:
                    self.rooms.append(room)
                    break
            else:
                raise ValueError(
                    "Room {} for statistical parameter {} not found.".format(
                        repr(room_name), repr(self.name)
                    )
                )
        if not self.cfg["rooms"]:
            # add all rooms if none are specified
            self.rooms.extend(self.app.rooms)


class AbstractValueCollectorMixin:
    """Abstract class for creating custom value collectors."""

    def collect_values(  # pylint: disable=no-self-use
        self,
    ) -> T.Iterable[T.Tuple[T.Any, T.Union[float, int]]]:
        """Should collect the implementation-specific values from which
        entries are built. The first item of each tuple is an identifier
        for the value (e.g. an entity id of the entity it originated
        from), the second is the value."""

        return []


class ActorValueCollectorMixin(AbstractValueCollectorMixin, RoomBasedParameter):
    """Helps collecting a value per actor."""

    def collect_values(self) -> T.Iterable[T.Tuple[T.Any, T.Union[float, int]]]:
        """Calls self.collect_actor_value() for all actors of the
        associated rooms and returns the collected values."""

        values = []
        for room in self.rooms:
            for actor in filter(lambda a: a.is_initialized, room.actors):
                value = self.collect_actor_value(  # pylint: disable=assignment-from-none
                    actor
                )
                self.log(
                    "Value for {} in {} is {}.".format(actor, room, value),
                    level="DEBUG",
                )
                if value is not None:
                    values.append((actor.entity_id, value))
        return values

    def collect_actor_value(  # pylint: disable=no-self-use,unused-argument
        self, actor: "ActorBase"
    ) -> T.Union[float, int, None]:
        """Should generate a value for the given actor or None, if the
        actor should be excluded."""

        return None

    def initialize_actor_listeners(
        self, actor: "ActorBase"
    ) -> None:  # pylint: disable=no-self-use,unused-argument
        """Should initialize the appropriate state listeners for the
        given actor."""

    def initialize_listeners(self) -> None:
        """Calls self.initialize_actor_listeners() for each actor of the
        associated rooms."""

        super().initialize_listeners()

        for room in self.rooms:
            for actor in filter(lambda a: a.is_initialized, room.actors):
                self.initialize_actor_listeners(actor)


class MinAvgMaxParameter(AbstractValueCollectorMixin, StatisticalParameter):
    """A generic parameter implementation that automatically calculates
    min/avg/max entries based on all values collected. Weighting of
    individual values is supported as well.
    It needs an implementation of AbstractValueCollectorMixin mixed in
    to have a source for values."""

    config_schema_dict = {
        **StatisticalParameter.config_schema_dict,
        vol.Optional("factors", default=dict): vol.All(
            lambda v: v or {},
            {
                util.CONF_STR_KEY: vol.All(
                    vol.Any(float, int), vol.Range(min=0, min_included=False)
                )
            },
        ),
        vol.Optional("weights", default=dict): vol.All(
            lambda v: v or {},
            {util.CONF_STR_KEY: vol.All(vol.Any(float, int), vol.Range(min=0))},
        ),
    }

    def generate_entries(self) -> T.Dict[str, StatisticalValueType]:
        """Generates min/avg/max parameter entries."""

        plain_values = self.collect_values()
        values = []
        weighted_sum = 0.0
        weights_sum = 0.0
        for _id, value in plain_values:
            weight = self.cfg["weights"].get(_id, 1)
            if weight == 0:
                continue

            value *= self.cfg["factors"].get(_id, 1)
            values.append(value)
            weighted_sum += weight * value
            weights_sum += weight

        _min = min(values) if values else 0.0
        _avg = weighted_sum / weights_sum if values else 0.0
        _max = max(values) if values else 0.0
        return {"min": _min, "avg": _avg, "max": _max}
